@@ -3,24 +3,27 @@
 # All Rights Reserved - See License
 #
 
-use v5.14;
-use utf8;
+use v5.8;
 
 package Parallel::WorkUnit;
-$Parallel::WorkUnit::VERSION = '1.001';
+$Parallel::WorkUnit::VERSION = '1.002';
 # ABSTRACT: Provide easy-to-use forking with ability to pass back data
 
 use strict;
 use warnings;
 use autodie;
 
+use TryCatch;
+my $do_thread = eval 'use threads qw//; 1' if $^O eq 'MSWin32';
+if ($do_thread) { eval 'use Thread::Queue;'; }
+
 use Carp;
+
 use IO::Pipe;
 use IO::Select;
 use Moose;
 use POSIX ':sys_wait_h';
 use Storable;
-use TryCatch;
 
 use namespace::autoclean;
 
@@ -32,11 +35,28 @@ has '_subprocs' => (
     init_arg => undef
 );
 
+# This only gets used on Win32.
+has '_queue' => (
+    is       => 'rw',
+    init_arg => undef
+);
+
+# This only gets used on Win32
+has '_count' => (
+    is       => 'rw',
+    isa      => 'Int',
+    init_arg => undef,
+    default  => 1
+);
+
 sub BUILD {
     my $self = shift;
 
     $self->_subprocs( {} );
-    $SIG{CHLD} = 'IGNORE';
+
+    if ($do_thread) {
+        $self->_queue( Thread::Queue->new() );
+    }
 }
 
 
@@ -46,35 +66,54 @@ sub async {
 
     my $pipe = IO::Pipe->new();
 
-    my $pid = fork;
-    if ( !defined($pid) ) { die "Fork failed: $!"; }
+    my ($pid, $thr);
+    if ($do_thread) {
+        $pid = $self->_count();
+        $self->_count($pid + 1);
+
+        $thr = threads->create( sub { $self->_child($sub, $pipe, $pid); } );
+        if ( !defined($thr) ) { die "thread creation failed: $!"; }
+    } else {
+        $pid = fork();
+    }
 
     if ($pid) {
-
         # We are in the parent process
         $pipe->reader();
 
         $self->_subprocs()->{$pid} = {
             fh       => $pipe,
-            callback => $callback
+            callback => $callback,
+            caller   => [ caller() ],
+            thread   => $thr
         };
 
         return $pid;
 
     } else {
+        $self->_child($sub, $pipe, undef);
+    }
+}
 
-        # We are in the child process
-        $pipe->writer();
-        $pipe->autoflush(1);
+sub _child {
+    if (scalar(@_) != 4) { confess 'invalid call'; }
+    my ($self, $sub, $pipe, $pid) = @_;
 
-        try {
-            my $result = $sub->();
-            $self->_send_result( $pipe, $result );
-        } catch($err) {
+    # We are in the child process
+    $pipe->writer();
+    $pipe->autoflush(1);
 
-            $self->_send_error( $pipe, $err );
-        };
+    try {
+        my $result = $sub->();
+        $self->_send_result( $pipe, $result, $pid );
+    } catch($err) {
+        $self->_send_error( $pipe, $err, $pid );
+    };
 
+    # Windows doesn't do fork(), it does threads...
+    if ($do_thread) {
+        return 1;
+    } else {
         exit();
     }
 }
@@ -87,25 +126,34 @@ sub waitall {
     my $sp = $self->_subprocs();
     if ( !keys(%$sp) ) { return; }
 
-    my $s = IO::Select->new();
-    foreach ( keys(%$sp) ) { $s->add( $sp->{$_}{fh} ); }
 
-    my @ready = $s->can_read;
+    if ($do_thread) {
+        my $child = $self->_queue()->dequeue(1);
 
-    foreach my $fh (@ready) {
-        foreach my $child ( keys(%$sp) ) {
-            if ( defined($fh->fileno())) {
-                if ( $fh->fileno() == $sp->{$child}{fh}->fileno() ) {
-                    $self->_read_result($child);
+        my $thr = $self->_subprocs()->{$child}{thread};
+        $self->_read_result($child);
+
+        $thr->join();
+    } else {
+        my $s = IO::Select->new();
+        foreach ( keys(%$sp) ) { $s->add( $sp->{$_}{fh} ); }
+
+        my @ready = $s->can_read();
+
+        foreach my $fh (@ready) {
+            foreach my $child ( keys(%$sp) ) {
+                if ( defined($fh->fileno())) {
+                    if ( $fh->fileno() == $sp->{$child}{fh}->fileno() ) {
+                        my $thr = $self->_subprocs()->{$child}{thread};
+                        $self->_read_result($child);
+
+                        waitpid($child, 0);
+                    }
                 }
             }
         }
     }
    
-    # Just in case SIG{CHLD}="IGNORE" doesn't cause the children to
-    # be reaped. 
-    while ( ( my $child = waitpid( -1, WNOHANG ) ) > 0 ) { }
-
     # Tail recursion
     goto &waitall;
 }
@@ -122,38 +170,49 @@ sub wait {
         return;
     }
 
+    my $thr = $self->_subprocs()->{$pid}{thread};
     my $result = $self->_read_result($pid);
 
-    # Just in case SIG{CHLD}="IGNORE" doesn't cause the children to
-    # be reaped. 
-    while ( ( my $child = waitpid( -1, WNOHANG ) ) > 0 ) { }
+    if ($do_thread) {
+        $thr->join();
+    } else {
+        waitpid($pid, 0);
+    }
 
     return $result;
 }
 
 sub _send_result {
-    if ( $#_ != 2 ) { confess 'invalid call'; }
-    my ( $self, $fh, $msg ) = @_;
+    if ( $#_ != 3 ) { confess 'invalid call'; }
+    my ( $self, $fh, $msg, $pid ) = @_;
 
-    $self->_send( $fh, 'RESULT', $msg );
+    $self->_send( $fh, 'RESULT', $msg, $pid );
 }
 
 sub _send_error {
-    if ( $#_ != 2 ) { confess 'invalid call'; }
-    my ( $self, $fh, $err ) = @_;
+    if ( $#_ != 3 ) { confess 'invalid call'; }
+    my ( $self, $fh, $err, $pid ) = @_;
 
-    $self->_send( $fh, 'ERROR', $err );
+    $self->_send( $fh, 'ERROR', $err, $pid );
 }
 
 sub _send {
-    if ( $#_ != 3 ) { confess 'invalid call'; }
-    my ( $self, $fh, $type, $data ) = @_;
+    if ( $#_ != 4 ) { confess 'invalid call'; }
+    my ( $self, $fh, $type, $data, $pid ) = @_;
+
+    my $msg = Storable::freeze( \$data );
+
+    if (!defined($msg)) {
+        die 'freeze() returned undef for child return value';
+    }
+
+    if ($do_thread) {
+        $self->_queue()->enqueue($pid);
+    }
 
     $fh->write($type);
     $fh->write("\n");
 
-    my $msg = Storable::freeze( \$data );
-
     $fh->write(length($msg));
     $fh->write("\n");
 
@@ -194,13 +253,17 @@ sub _read_result {
 
     my $data = ${ Storable::thaw($result) };
 
+    my $caller = $self->_subprocs()->{$child}{caller};
+    my $thr = $self->_subprocs()->{$child}{thread};
     delete $self->_subprocs()->{$child};
     $fh->close();
 
     if ( $type eq 'RESULT' ) {
         $cinfo->{callback}->($data);
     } else {
-        die("Child died with error: $data");
+        if ($do_thread) { $thr->join(); }
+        die("Child (created at " . $caller->[1] . " line " . $caller->[2] .
+            ") died with error: $data");
     }
 }
 
@@ -220,7 +283,7 @@ Parallel::WorkUnit - Provide easy-to-use forking with ability to pass back data
 
 =head1 VERSION
 
-version 1.001
+version 1.002
 
 =head1 SYNOPSIS
 
@@ -270,6 +333,11 @@ die (inside the C<waitall()> method).
 The PID of the child is returned to the parent process when
 this method is executed.
 
+Note: on Windows with threaded Perl, threads instead of forks are used.
+See C<thread> for the caveats that apply.  The PID returned is instead
+a meaningless (outside of this module) counter, not associated with any
+Windows thread identifier.
+
 =head2 waitall()
 
 Called from the parent method while waiting for the children
@@ -280,10 +348,6 @@ method will return.
 If a child dies unexpectedly, this method will C<die()> and propagate a
 modified exception.
 
-This method has a side-effect of reaping any children that haven't yet
-been reaped.  If you aren't messing with C<$SIG{CHLD}>, you do not need
-to worry about this.
-
 =head2 wait($pid)
 
 This functions simiarly to C<waitall()>, but only waits for
@@ -295,17 +359,21 @@ executing, it simply returns.  Otherwise, it waits until the
 child process's work unit is complete and executes the callback
 routine, then returns.
 
-This method has a side-effect of reaping any children that haven't yet
-been reaped.  If you aren't messing with C<$SIG{CHLD}>, you do not need
-to worry about this.
-
 =head1 SIDE EFFECTS
 
-This module, when instantiated, sets the SIG{CHLD} handler to 'IGNORE'.
-This means that children will be automatically reaped on most systems.
 In addition, the C<wait()> and C<waitall()> methods will attempt to
 reap any outstanding children.
 
+=head1 BUGS
+
+Windows doesn't do C<fork()>, but emulates it with threads.  As a result,
+any thread unsafe library is going to cause problems with Windows.  In
+addition, all the normal thread caveats apply - see L<threads> for more
+information.
+
+In addition, this code is unlikely to function properly on a Windows without
+threaded Perl.
+
 =head1 AUTHOR
 
 Joel Maslak <jmaslak@antelope.net>
